#--
# Copyright (c) 2011-2016 Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#++

require "bughunting/common"
require "bughunting/client/config"
require "bughunting/client/error"
require "bughunting/common/task"
require "bughunting/common/local_execute"
require "pathname"
require "xmlrpc/client"
require 'fileutils'

class Array
  def check val
    self.any? do
      |line|
      line.include? val
    end
  end
end

class ClientController

  include LocalExecute

  attr_accessor :client_config

  def initialize(client_config_file, user_key, task_name_i, verbose = false)
    @client_config = ClientConfig.load_file client_config_file
    @user_key = user_key
    @task_name = task_name_i
    @xmlrpc = XMLRPC::Client.new3({
      "host" => @client_config["server.host"],
      "port" => @client_config["server.port"],
      "timeout" => @client_config["server.timeout"]
    })
    @xmlrpc.http_header_extra = { "accept-encoding" => "identity" }
    @verbose = verbose
  end

  def task_name
    @task_name ||= autodetect_task!
  end

  def check
    $stderr.puts "Submitting task '%s'." % task_name
    (success, message) = @xmlrpc.call "bughunting.check", @user_key, task_name, create_patch
    print_check_result success, message
    success
  end

  def git_check
#    unless Dir.chdir @client_config['paths']['sources_dir']
#      raise ClientError.new \
#        "Fatal error: Cannot chdir to sources dir: #{@client_config['paths']['sources_dir']}."
#    end

    # Let's check the remote
    unless run("git remote -v") .check "origin\tgit@gitlab.com:summer-camp-#{Time.new.year}/"
      $stderr.puts "Invalid git remote, please ask for help."
      return false
    end

    # Let's retry twice, to be sure
    3.times do
      run "git fetch origin"

      status = run "git status -uno"

      debug status

      break if status.check "modified:"
      break if status.check "Changes not staged for commit:"
      break if status.check "Untracked files:"

      unless status.check "On branch main"
        result = run "git checkout main"

        if result.check "error: "
          $stderr.puts "Error when switching to main branch.\n" \
            + "run `git status` for more."

          break
        end

        next
      end

      unless ENV["HUNT_IGNORE_GIT"]
        break unless status.check \
          "Your branch is up to date with 'origin/main'."
      end

      break unless status.check "nothing to commit"

      return true
    end

    false
  end

  def clean
    $stderr.puts "Cleaning task '%s'" % task_name
    task_cache_sources = File.join(@client_config['paths']['cache_dir'], task_name, 'source')
    task_path_to_clean = File.join(File.expand_path(@client_config['paths']['sources_dir']), task_name)

    debug
    FileUtils.rm_rf task_path_to_clean
    FileUtils.cp_r task_cache_sources, task_path_to_clean
  end

  def serverclean
    $stderr.puts "Cleaning task #{task_name}"
    @xmlrpc.call 'bughunting.clean', @user_key, task_name
  end

  def list
    # ignore task_name
    task_list = @xmlrpc.call "bughunting.list_tasks", @user_key

    columns = `stty size 2>/dev/null`.split[1].to_i
    columns = 120 if columns.nil? or columns < 120

    name_max = 0
    task_list.each do |task_info|
      length = task_info["name"].length
      name_max = length if length > name_max
    end

    summary_max = columns - name_max - 7

    task_list.each do |task_info|
      name = ("%-#{name_max}s" % task_info["name"]).bold
      summary = task_info["summary"][0..summary_max]
      summary += "..." if summary.length != task_info["summary"].length

      puts "%s - %s" % [name, summary]
    end

    true
  end

  def patch
    puts create_patch
  end

  def create_patch
    cache_dir = File.join @client_config['paths']['cache_dir'], task_name
    task = Task.new cache_dir

    patches = task.config['check_files'].inject([]) do |patch, target_name|
      target_path = File.join @client_config['paths']['sources_dir'], task_name
      cache_dir_sources = File.join cache_dir, 'source'
      original_file = if task.config['copy_task_on_server']
                        File.join(cache_dir_sources, target_name)
                      else
                        '/dev/null'
                      end

      create_diff(original_file, target_name, target_path, patch)
    end

    # Multiple empty patches will break the `patch` on the server.
    patches.reject(&:empty?)
  end

  private

  # Actually runs `diff` on two specified files in specified directory
  # and appends the output into `patch`.
  def create_diff(original_file, target_file, target_path, patch)
    local_execute(
      [
        "diff", "-Naur",
        original_file,
        target_file
      ],
      work_dir: target_path,
      out: patch,
      debug: @verbose
    )
    patch
  end

  def relative_path(origin_path, target_path)
    Pathname.new(target_path).relative_path_from(Pathname.new(origin_path)).to_s
  end

  def autodetect_task!
    cwd = `/bin/pwd -L`.strip
    sources = File.join(File.expand_path(@client_config['paths']['sources_dir']), "/")

    if cwd.start_with? sources
      cwd[sources.length .. -1].split("/").first
    else
      raise ClientError.new "Cannot autodetect the task name. Please, specify manually."
    end
  end

  def print_check_result(success, message)
    $stderr.puts "%s: %s" % [
      success ? "OK".bold.green : "ERROR".bold.red,
      message
    ]
  end

  def debug msg = '', &block
    if @verbose || ( ENV["DEBUG"] && !ENV["DEBUG"].empty? )
      puts msg
      yield if block_given?
    end
  end

  def run command
    debug "Running '#{command}'"

    ret = %x{#{command} 2>&1} .split(?\n) - [""]
  end
end
